#======================================================================
# ZMapper - use a delta Z map (matrix) or Barycentric interpolation
# to adjust Z movements for each extrusion move and insert the Z into 
# each line of code. This is a Z compensation to correct for minor
# mechanical and calibration issues on a delta printer. You must
# calibrate the delta radius, tower lengths and Z=0 before applying
# this script.
#
# Before: G1 X-4.24 Y29.08  E13.124
# After:  G1 X-4.24 Y29.08 Z34.4 E13.124
#
# The Z is the current layer height in mm plus the delta Z value at
# that X-Y position.
#
# USAGE: ZMapper.py -i gcodefile.gcode -n -t <trianglefile> -z heightmapfile.csv [-q]
#
# the -i file must have a ".gcode" extension
# the output file will be renamed as inptfilename-ZMapped.gcode
#
# only one of -t or -z is needed.
#
# -q prevents the postprocessor from running, this way you can disable
# the script without deleting the postprocessor line.
#
# Copyright (C) 2015 Michael Hackney
#======================================================================
import array
import sys
import os
import string
import csv
import getopt
import math

#---------------------------------------------------------------------------------------------
# Test if pointT is inside the triangle described by pointA, pointB and pointC using
# Barycentric coordinates. If it is, simply calculate the deltaZ for the test point
# using Barycentric interpolation of the pre-calculated Barycentric coordinates.
#
def deltaZfromTriangle(pointT, triangle):
    
    pointA = triangle[0]
    pointB = triangle[1]
    pointC = triangle[2]
    
    denomintor = (((pointB[1] - pointC[1]) * (pointA[0] - pointC[0])) + ((pointC[0] - pointB[0]) * (pointA[1] - pointC[1])))
                
    a = (((pointB[1] - pointC[1]) * (pointT[0] - pointC[0])) + ((pointC[0] - pointB[0]) * (pointT[1] - pointC[1]))) / denomintor
    b = (((pointC[1] - pointA[1]) * (pointT[0] - pointC[0])) + ((pointA[0] - pointC[0]) * (pointT[1] - pointC[1]))) / denomintor
    c = 1 - a - b

    if ((0 <= a) and (a <= 1) and (0 <= b) and (b <= 1) and (0 <= c) and (c <= 1)):
        return (a * pointA[2]) + (b * pointB[2]) + (c * pointC[2]) 
    else:
        return "not found"        

#---------------------------------------------------------------------------------------------
# basez is the current slice height, x and y are the postions from the center of the bed (0,0)
#
def getZ(basez, x, y):
    X = x/10 + 14
    Y = y/10 + 14
    
    for row in heightMap:
        for e in row:
            print '%1.2f' % e,
        print "\n"
        
    print X
    print Y
    Z = basez + heightMap[X][Y]
    print Z

#---------------------------------------------------------------------------------------
# determine if the line is a begin layer section, if so, return the layer height (in mm)    
#
def find_begin_layer(line):
    if line.startswith('; BEGIN_LAYER_OBJECT z='):
        zIndex = line.find('z=')
        return line[zIndex+2:]
    return False

#-----------------------------------------------------------------------------------------------------
# determine if the line is a G1 move. If so, make sure it isn't a prime or Zlift (return False if so).
# If it's a normal extrusion move, return the line substrings.
#
def find_extrude_segment(line):
    # only process lines that start with a G1 movement
    if line.startswith('G1 X'):
        # make sure it isnt a prime or Z lift movement
        if 'Z' in line:
            return False
        lineParts = line.split()
        return lineParts

#----------------------------------------------------------------------------------------
# Use the X and Y positions in lineParts to lookup the Z height adjustment, add that to
# layerHeight and return a formatted gcode line. Currently supports two types of mapping:
#    1) a matrix of delta Z values
#    2) Barymetric Interpolation over a set of triangles   
#
def mapZ(lineParts, layerHeight, interpolationType):
    X = 0
    Y = 0
    
    # perform Barycentric Interpolation
    if interpolationType == 1:

        # find X, Y for the move
        for part in lineParts:
            firstChar = part[0]
        
            if firstChar == 'X':   
                X = float(X)
            
            elif firstChar == 'Y':
                Y = float(part[1:])

        if ((X >= 0) and (Y >= 0)):
            for triangle in NEquad:
                deltaZ = deltaZfromTriangle([X,Y], triangle)
                if deltaZ != "not found":
                    break
    
        elif ((X >= 0) and (Y < 0)):
            for triangle in SEquad:
                deltaZ = deltaZfromTriangle([X,Y], triangle)
                if deltaZ != "not found":
                    break

        elif ((X < 0) and (Y < 0)):
            for triangle in SWquad:
                deltaZ = deltaZfromTriangle([X,Y], triangle)
                if deltaZ != "not found":
                   break
            
        elif ((X < 0) and (Y < 0)):
            for triangle in NWquad:
                deltaZ = deltaZfromTriangle([X,Y], triangle)
                if deltaZ != "not found":
                    break       

    # use matrix lookup
    else:
        # Here's where I'd like to do a linear interpolation to get a better Z estimate.
        #

        # find X, Y for the move and divide it by 10 for the 10mmx10mm cell size
        for part in lineParts:
            firstChar = part[0]
        
            if firstChar == 'X':   
                X = int(round(float(X)/10) + columns/2)
            
            elif firstChar == 'Y':
                Y = part[1:]
                Y = int(round(float(Y)/10) + rows/2)

        deltaZ = heightMap[X][Y]
        
    newZ = float(deltaZ) + float(layerHeight)
    zLine = 'Z' + str(newZ)

    lineParts.insert(3, zLine.strip())
    newCodeLine = string.join(lineParts, " ")

    return newCodeLine

#--------------------------------------------------------------------------
# Read the delta Z matrix from a CSV file and create a two dimensional list
#
def makeDeltaZ(heightmap):    
    with open(heightmap) as file:
        reader = csv.reader(file)
        htmap = list(reader)
        return htmap
                           
#====================================================
# __main__
#====================================================
if __name__ == '__main__':

    global NEquad
    global SEquad
    global SWquad
    global NWquad
    global heightMap

    try:
        opts, args = getopt.getopt(sys.argv[1:],"i:t:z:q")
    except getopt.GetoptError as err:
        print str(err)
        sys.exit('usage: ZMapper.py -i <gcodefile> -t <trianglefile> -z <heightmapfile> [-q]')
    
    interpolationType = 0
    
    for o, a in opts:
        if o == "-q":
            sys.exit()
        elif o == "-i":
            gcodefn = a
        elif o == "-t":
            trianglefn = a
            interpolationType = 1
        elif o == "-z":
            heightfn = a
            interpolationType = 2
    
    if interpolationType == 1:
        # triangles are 3 points with the delta Z height at each point as:
        # p1[X1, Y1, Z1], p2[X2, Y2, Z2], p3[X3, Y3, Z3]
        
        #=========================================================================
        # EDIT THIS SECTION TO INCLUDE YOUR PROBE DATA 
        # by editing the last value (0.0) in the "[]" with your probe data
        # OR copy and paste the automatically generated code from the green
        # cells in the accompanying spreadsheet
        #=========================================================================
                
        # center
        pt1 = [0,0,0.0]

        # outer ring
        pt2 = [-110,-60,0.0]
        pt3 = [-60,-110,0.0]
        pt4 = [0,-130,0.0]
        pt5 = [60,-110,0.0]
        pt6 = [110,-60,0.0]
        pt7 = [130,0,0.0]
        pt8 = [110,60,0.0]
        pt9 = [60,110,0.0]
        pt10 = [0,130,0.0]
        pt11 = [-60,110,0.0]
        pt12 = [-110,60,0.0]
        pt13 = [-130,0,0.0]

        # middle ring
        pt14 = [-70,-40,0.0]
        pt15 = [0,-80,0.0]
        pt16 = [70,-40,0.0]
        pt17 = [70,40,0.0]
        pt18 = [0,80,0.0]
        pt19 = [-70,40,0.0]

        # inner ring
        pt20 = [-20,30,0.0]
        pt21 = [-40,0,0.0]
        pt22 = [-20,-30,0.0]
        pt23 = [20,-30,0.0]
        pt24 = [40,0,0.0]
        pt25 = [20,30,0.0]
        
        # outer ring triangles
        t1 = [pt10,pt11,pt18,"t1"]
        t2 = [pt11,pt18,pt19,"t2"]
        t3 = [pt11,pt12,pt19,"t3"]
        t4 = [pt12,pt13,pt19,"t4"]
        t5 = [pt13,pt14,pt19,"t5"]
        t6 = [pt2,pt13,pt14,"t6"]
        t7 = [pt2,pt3,pt14,"t7"]
        t8 = [pt3,pt14,pt15,"t8"]
        t9 = [pt3,pt4,pt15,"t9"]
        t10 = [pt4,pt5,pt15,"t10"]
        t11 = [pt5,pt15,pt16,"t11"]
        t12 = [pt5,pt6,pt16,"t12"]
        t13 = [pt6,pt7,pt16,"t13"]
        t14 = [pt7,pt16,pt17,"t14"]
        t15 = [pt7,pt8,pt17,"t15"]
        t16 = [pt8,pt9,pt17,"t16"]
        t17 = [pt9,pt17,pt18,"t17"]
        t18 = [pt9,pt10,pt18,"t18"]

        # middle ring triangles
        t19 = [pt18,pt19,pt20,"t19"]
        t20 = [pt19,pt20,pt21,"t20"]
        t21 = [pt14,pt19,pt21,"t21"]
        t22 = [pt14,pt21,pt22,"t22"]
        t23 = [pt14,pt15,pt22,"t22"]
        t24 = [pt15,pt22,pt23,"t24"]
        t25 = [pt15,pt16,pt23,"t25"]
        t26 = [pt16,pt23,pt24,"t26"]
        t27 = [pt16,pt17,pt24,"t27"]
        t28 = [pt17,pt24,pt25,"t28"]
        t29 = [pt17,pt18,pt25,"t29"]
        t30 = [pt18,pt20,pt25,"t30"]

        # inner ring triangles
        t31 = [pt1,pt20,pt25,"t31"]
        t32 = [pt1,pt20,pt21,"t32"]
        t33 = [pt1,pt21,pt22,"t33"]
        t34 = [pt1,pt22,pt23,"t34"]
        t35 = [pt1,pt23,pt24,"t35"]
        t36 = [pt1,pt24,pt25,"t36"]
    
        NEquad = [t31, t36, t30, t29, t28, t27, t14, t15, t16, t17, t18]
        SEquad = [t34, t35, t24, t26, t25, t27, t14, t13, t12, t11, t10]
        SWquad = [t34, t33, t24, t23, t22, t21, t5,  t6,  t7,  t8,  t9]
        NWquad = [t31, t32, t21, t20, t19, t30, t1,  t2,  t3,  t4,  t5]
        
        #=========================================================================
        # END OF EDIT SECTION
        #=========================================================================
                
    elif interpolationType == 2:
        heightMap = makeDeltaZ(str(heightfn))
        columns = len(heightMap)
        rows = len(heightMap[:])
    
    else:
        print "you must specify either a triangle or heightmap file"
    
    # Now let's map the Zs
    #
    infile = open(gcodefn, 'rb')
    currentZ = 0
     
    newfn = string.replace(gcodefn, ".gcode","-ZMapped.gcode")
    with open(newfn, 'w') as outfile:
        outfile.write(';---------------------------------------------------------------')
        outfile.write("\n")    
        outfile.write('; Processed with ZMapper using ')
        if interpolationType == 1:
            outfile.write('Barycentric Interpolation')
        else:
            outfile.write('Height Map Matrix')
        
        outfile.write("\n")
        outfile.write(';---------------------------------------------------------------')
        outfile.write("\n")    
        
        for line in infile:
            layerHeight = find_begin_layer(line)

            if layerHeight != False:
                currentZ = layerHeight

            lineParts = find_extrude_segment(line)
            if lineParts != False:
                if lineParts != None:
                    mappedLine = mapZ(lineParts, currentZ, interpolationType)
                    outfile.write(mappedLine)
                    outfile.write("\n")
                else:
                    outfile.write(line.replace('\r',''))
            
        infile.close()
        outfile.close()